-- misc outputs
library ieee;
use ieee.std_logic_1164.all;
library exemplar;
use exemplar.exemplar_1164.all;
use std.textio.all;

entity outputs is
    port (
		cpu_adr : out std_logic_vector(4 downto 1);
		cpu_data : out std_logic_vector(15 downto 0);
		cpu_not_cs,cpu_rw : out std_logic;
		pbus : in std_logic_vector(15 downto 0);
		coefd : out std_logic_vector(6 downto 0);
		qsf_out : out std_logic_vector(1 downto 0);
		hen_in : out std_logic_vector(3 downto 0)
		   );
end outputs;

library ieee;
use ieee.std_logic_1164.all;
library exemplar;
use exemplar.exemplar_1164.all;
use std.textio.all;

architecture operate_outputs of outputs is

type memory is array (integer range <>) of integer;


procedure cpuwr(
  file romdef : TEXT;
  adr,data,timev : inout memory;
  flength : inout integer 
  ) is
variable l : line;
variable j : integer;
variable chr : character;
variable sptr : integer :=0;
variable forever : integer :=16000000;
variable flag : boolean;
variable adrv,datav,timeval : integer;


        type hex_to_v4_t is array(character) of integer;
        constant hex_to_v4 : hex_to_v4_t := (
            '0' => 0,
            '1' => 1,
            '2' => 2,
            '3' => 3,
            '4' => 4,
            '5' => 5,
            '6' => 6,
            '7' => 7,
            '8' => 8,
            '9' => 9,
            'a' | 'A' => 10,
            'b' | 'B' => 11,
            'c' | 'C' => 12,
            'd' | 'D' => 13,
            'e' | 'E' => 14,
            'f' | 'F' => 15,
            others => -1);

        -------------------------------------------
        -- convert a hex string to integer
        -------------------------------------------
        procedure hex_str_to_v( s : in string; ptr : inout integer; result : inout integer:=0; flag : inout boolean)  is
            variable mult : integer :=1;
        begin
			flag := false;
			result :=0;
            for i in ptr to s'high loop
			    ptr := i;
			    if ((hex_to_v4(s(i)) = -1) and (flag)) then exit; end if; -- hit terminator
                result := result*16;
			    if (hex_to_v4(s(i)) /= -1) then
				  flag := true;
                  result := result + hex_to_v4(s(i));
				end if;
            end loop;
        end procedure;

        type dec_to_v4_t is array(character) of integer;
        constant dec_to_v4 : dec_to_v4_t := (
            '0' => 0,
            '1' => 1,
            '2' => 2,
            '3' => 3,
            '4' => 4,
            '5' => 5,
            '6' => 6,
            '7' => 7,
            '8' => 8,
            '9' => 9,
            others => -1);

        -------------------------------------------
        -- convert a decimal string to integer
        -------------------------------------------
        procedure dec_str_to_v( s : in string; ptr : inout integer; result : inout integer:=0; flag : inout boolean)  is
            variable mult : integer :=1;
        begin
			flag := false;
			result :=0;
            for i in ptr to s'high loop
			    ptr := i;
			    if ((dec_to_v4(s(i)) = -1) and (flag)) then exit; end if; -- hit terminator
                result := result*10;
			    if (dec_to_v4(s(i)) /= -1) then
				  flag := true;
                  result := result + dec_to_v4(s(i));
				end if;
            end loop;
        end procedure;
begin
  for j in 0 to forever loop
    readline(romdef,l);
	sptr :=1;
    hex_str_to_v(l.all,sptr,adrv,flag);
    hex_str_to_v(l.all,sptr,datav,flag);
    dec_str_to_v(l.all,sptr,timeval,flag);
	adr(j) := adrv;
	data(j) := datav;
	timev(j) := timeval;
	flength := j+1;
    if endfile(romdef) then exit; end if;
  end loop;
end cpuwr;


procedure f_rominit(rom : inout memory; file romdef : TEXT) is
variable l : line;
variable j : integer;
variable chr : character;
        type hex_to_v4_t is array(character) of integer;
        constant hex_to_v4 : hex_to_v4_t := (
            '0' => 0,
            '1' => 1,
            '2' => 2,
            '3' => 3,
            '4' => 4,
            '5' => 5,
            '6' => 6,
            '7' => 7,
            '8' => 8,
            '9' => 9,
            'a' | 'A' => 10,
            'b' | 'B' => 11,
            'c' | 'C' => 12,
            'd' | 'D' => 13,
            'e' | 'E' => 14,
            'f' | 'F' => 15,
            others => 0);

        -------------------------------------------
        -- convert a hex string to integer
        -------------------------------------------
        function hex_line_to_v( s : string) return integer is
            variable mult : integer :=1;
            variable result : integer :=0;
        begin
            for i in s'reverse_range loop
                result := result + (hex_to_v4(s(i))*mult);
                mult := mult*16;
            end loop;
            return result;
        end function;
begin
  for j in rom'range loop
    readline(romdef,l);
    if endfile(romdef) then exit; end if;
    rom(j) := hex_line_to_v(l.all);
  end loop;
end f_rominit;

constant HEN1STR  : string := "hen1.tab";
FILE HEN1T : TEXT open read_mode is HEN1STR;  
constant hen_time : time := 37 ns;

constant qsfstr  : string := "qsf.tab";
FILE qsft : TEXT open read_mode is qsfstr;  
constant qsf_time : time := 8000 ns;

constant coefstr  : string := "coef.tab";
FILE coeft : TEXT open read_mode is coefstr;  
constant coef_time : time := 8000 ns;

constant cpustr  : string := "cpu.tab";
FILE cpufile : TEXT open read_mode is cpustr;  

begin


proc_hen1 : process	
variable HEN1 : memory(1 to 16);
variable hi : integer :=1;
variable forever : integer :=16000000;
begin
  f_rominit(HEN1,HEN1T);
  for i in 0 to forever loop
	  hen_in <= int2evec(HEN1(hi),4);
	  hi := hi+1;
      if hi>HEN1'high then
        hi := 1;
      end if;
    wait for hen_time;
  end loop;
end process;
  
proc_qsf : process	
variable qsf : memory(1 to 4);
variable hi : integer :=1;
variable forever : integer :=16000000;
begin
  f_rominit(qsf,qsft);
  for k in 0 to forever loop
	  qsf_out <= int2evec(qsf(hi),2);
	  hi := hi+1;
      if hi>qsf'high then
        hi := 1;
      end if;
    wait for qsf_time;
  end loop;
end process;

proc_coef : process	(pbus)
variable coef : memory(0 to 65535);
variable startup : boolean :=true;
begin
  if startup then
    f_rominit(coef,coeft);
    startup := false;
  end if;
	  coefd <= int2evec(coef(evec2int(pbus)))(6 downto 0) after 150 ns;

end process;

proc_cpu: process
variable adr,data,timev : memory(0 to 30000);
variable timex : integer;
variable timeval : time;
variable flength : integer;
variable startup : boolean :=true;

constant start_cs : time :=50 ns;
constant cs_time : time :=250 ns;
begin
  if startup then
	cpuwr(cpufile,adr,data,timev,flength);
	cpu_not_cs <= '1';
  for i in 0 to flength-1	loop
    cpu_adr <= int2evec(adr(i),4);
	cpu_data <= int2evec(data(i),16);
	cpu_rw <= '0';
	cpu_not_cs <= TRANSPORT '0' after start_cs;
	cpu_not_cs <= TRANSPORT '1' after cs_time;
	timex := timev(i);
	timeval := time'(timex* 1 ns);
	wait for timeval;
  end loop;
    startup := false;
  end if;
  wait for 1 ns;
end process;

end operate_outputs;

